@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2020 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2022 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -39,6 +39,7 @@
 #ifdef WIN32
  #include <winsock.h>
 #else /* UNIX */
+ #include <signal.h>
  #include <sys/types.h>
  #include <time.h>
  #include <unistd.h>
@@ -125,40 +126,42 @@ extern char** environ;
 #define WRAPPER_JSTATE_DOWN_CHECK 70 /* JVM is confirmed to be down, but we still need
                                      *  to do our cleanup work.  This is the state after
                                      *  a JVM process has gone away. */
-#define WRAPPER_JSTATE_DOWN_FLUSH 71 /* JVM is confirmed to be down but we still need
+#define WRAPPER_JSTATE_DOWN_FLUSH_STDIN 71 /* JVM is confirmed to be down but we still need
+                                            *  flush all bytes in stdin. */
+#define WRAPPER_JSTATE_DOWN_FLUSH 72 /* JVM is confirmed to be down but we still need
                                      *  flush and process all of its output. */
-#define WRAPPER_JSTATE_DOWN_CLEAN 72 /* JVM is confirmed to be down and we have cleaned
+#define WRAPPER_JSTATE_DOWN_CLEAN 73 /* JVM is confirmed to be down and we have cleaned
                                      *  up and flushed all output.  This is the initial
                                      *  state and the state after the JVM process has
                                      *  gone away and and cleanup has been done. */
-#define WRAPPER_JSTATE_LAUNCH_DELAY 73 /* Set from the DOWN state to launch a JVM.  The
+#define WRAPPER_JSTATE_LAUNCH_DELAY 74 /* Set from the DOWN state to launch a JVM.  The
                                      *  timeout will be the time to actually launch
                                      *  the JVM after any required delay. */
-#define WRAPPER_JSTATE_RESTART   74 /* JVM is about to be restarted. No timeout. */
-#define WRAPPER_JSTATE_LAUNCH    75 /* JVM is about to launch a JVM. No timeout. */
-#define WRAPPER_JSTATE_LAUNCHING 76 /* JVM was launched, but has not yet responded.
+#define WRAPPER_JSTATE_RESTART   75 /* JVM is about to be restarted. No timeout. */
+#define WRAPPER_JSTATE_LAUNCH    76 /* JVM is about to launch a JVM. No timeout. */
+#define WRAPPER_JSTATE_LAUNCHING 77 /* JVM was launched, but has not yet responded.
                                      *  Must enter the LAUNCHED state before <t>
                                      *  or the JVM will be killed. */
-#define WRAPPER_JSTATE_LAUNCHED  77 /* JVM was launched, and responed to a ping. */
-#define WRAPPER_JSTATE_STARTING  78 /* JVM has been asked to start.  Must enter the
+#define WRAPPER_JSTATE_LAUNCHED  78 /* JVM was launched, and responed to a ping. */
+#define WRAPPER_JSTATE_STARTING  79 /* JVM has been asked to start.  Must enter the
                                      *  STARTED state before <t> or the JVM will be
                                      *  killed. */
-#define WRAPPER_JSTATE_STARTED   79 /* JVM has responded that it is running.  Must
+#define WRAPPER_JSTATE_STARTED   80 /* JVM has responded that it is running.  Must
                                      *  respond to a ping by <t> or the JVM will
                                      *  be killed. */
-#define WRAPPER_JSTATE_STOP      80 /* JVM is about to be sent a stop command to shutdown
+#define WRAPPER_JSTATE_STOP      81 /* JVM is about to be sent a stop command to shutdown
                                      *  cleanly. */
-#define WRAPPER_JSTATE_STOPPING  81 /* JVM was sent a stop command, but has not yet
+#define WRAPPER_JSTATE_STOPPING  82 /* JVM was sent a stop command, but has not yet
                                      *  responded.  Must enter the STOPPED state
                                      *  and exit before <t> or the JVM will be killed. */
-#define WRAPPER_JSTATE_STOPPED   82 /* JVM has responed that it is stopped. */
-#define WRAPPER_JSTATE_KILLING   83 /* The Wrapper is about ready to kill the JVM
+#define WRAPPER_JSTATE_STOPPED   83 /* JVM has responed that it is stopped. */
+#define WRAPPER_JSTATE_KILLING   84 /* The Wrapper is about ready to kill the JVM
                                      *  process but it must wait a few moments before
                                      *  actually doing so.  After <t> has expired, the
                                      *  JVM will be killed and we will enter the STOPPED
                                      *  state. */
-#define WRAPPER_JSTATE_KILL      84 /* The Wrapper is about ready to kill the JVM process. */
-#define WRAPPER_JSTATE_KILLED    85 /* The Wrapper has requested termination of the JVM and 
+#define WRAPPER_JSTATE_KILL      85 /* The Wrapper is about ready to kill the JVM process. */
+#define WRAPPER_JSTATE_KILLED    86 /* The Wrapper has requested termination of the JVM and 
                                        confirmed the JVM is actually killed */
 
 /* Defined Action types.  Registered actions are negative.  Custom types are positive. */
@@ -249,6 +252,7 @@ struct WrapperConfig {
     TCHAR   *argCommand;            /* The command used to launch the wrapper. */
     TCHAR   *argCommandArg;         /* The argument to the command used to launch the wrapper. */
     TCHAR   *argConfFile;           /* The name of the config file from the command line. */
+    int     confFileOptional;       /* Whether the conf file is required or not. */
     TCHAR   *confDir;
     int     argConfFileDefault;     /* True if the config file was not specified. */
     int     argConfFileFound;       /* True if the config file was found. */
@@ -271,7 +275,7 @@ struct WrapperConfig {
     int     timerSlowThreshold;     /* If the difference between the system time based tick count and the timer tick count ever grows by more than this value then a warning will be displayed. */
     int     useTickMutex;           /* TRUE if access to the tick count should be protected by a mutex. */
     int     uptimeFlipped;          /* TRUE when the maximum uptime has been flipped. (Overflown) */
-
+    int     runCommonStarted;       /* TRUE once wrapperRunCommon() reported the 'started' status (used to know when to print the 'stop' status). */
     int     ignoreSequenceGaps;     /* TRUE if all sequence properties should be used. */
     int     port;                   /* Port number which the Wrapper is configured to be listening on */
     int     portMin;                /* Minimum port to use when looking for a port to bind to. */
@@ -287,6 +291,7 @@ struct WrapperConfig {
     TCHAR   *configFile;            /* Name of the configuration file */
     int     commandLogLevel;        /* The log level to use when logging the java command. */
     int     printJVMVersion;        /* tells the Wrapper to create a temp JVM to query the version, before starting the java application */
+    int     jvmVersionFailed;       /* Flag which is set to true when the forked process used to print the Java version exited with an error. */
     JavaVersion *javaVersion;       /* Java version of the current or next JVM instance to be launched. */
     JavaVersion *javaVersionMin;    /* The minimum version of Java required for the Wrapper to launch a JVM. */
     JavaVersion *javaVersionMax;    /* The maximum version of Java required for the Wrapper to launch a JVM. */
@@ -302,6 +307,7 @@ struct WrapperConfig {
     TCHAR   **jvmVersionCommand;    /* Command used to launch the JVM and request its version */
     TCHAR   **jvmCommand;           /* Command used to launch the JVM */
 #endif
+    int     javaCommandNotBinary;   /* Flag used to warn when the Java command is not a binary. */
     int     detachStarted;          /* TRUE if the JVM process should be detached once it has started. */
     int     environmentClasspath;   /* TRUE if the classpath should be passed to the JVM in the environment. */
     TCHAR   *classpath;             /* Classpath to pass to the JVM. */
@@ -310,6 +316,7 @@ struct WrapperConfig {
     TCHAR   key[17];                /* Key which the JVM uses to authorize connections. (16 digits + \0) */
     int     isConsole;              /* TRUE if the wrapper was launched as a console. */
     int     cpuTimeout;             /* Number of seconds without CPU before the JVM will issue a warning and extend timeouts */
+    int     pidLogLevel;            /* The log level to use when logging the PID of the JVM after it is launched. */
     int     startupTimeout;         /* Number of seconds the wrapper will wait for a JVM to startup */
     int     pingTimeout;            /* Number of seconds the wrapper will wait for a JVM to reply to a ping */
     int     pingAlertThreshold;     /* Number of seconds without a ping response that the Wrapper will start to warn about a slow ping. */
@@ -325,7 +332,14 @@ struct WrapperConfig {
     int     jvmSilentKill;          /* TRUE if the JVM should be silently killed at the next opportunity. */
 #ifdef WIN32
     int     javaIOBufferSize;       /* Size of the pipe buffer to use for java I/O. */
+#else
+    int     javaINBufferSize;       /* Size of the pipe buffer to use for java input. */
+    int     javaINFlush;            /* Flag to specify whether the JavaIN thread should flush stdin or not on a JVM restart. */
+    int     javaINFlushed;          /* Flag to inform the main event loop that the JavaIN thread finished to flush stdin. */
+    int     javaINFlushing;         /* Flag to inform the main event loop that the JavaIN thread is currently flushing. */
+    int     disableConsoleInputPermanent;   /* TRUE if console input should be permanantly disabled (will force disableConsoleInput to be TRUE on configuation reload). */
 #endif
+    int     disableConsoleInput;    /* TRUE if console input should be disabled. */
     int     useJavaIOThread;        /* If TRUE then a dedicated thread will be used to process console output form the JVM. */
     int     pauseThreadMain;        /* Number of seconds to pause the main thread on its next loop.  Only used for testing. */
     int     pauseThreadTimer;       /* Number of seconds to pause the timer thread on its next loop.  Only used for testing. */
@@ -339,10 +353,12 @@ struct WrapperConfig {
     DWORD   javaPID;                /* PID of the Java process. */
     HANDLE  wrapperProcess;         /* Handle of the Wrapper process. */
     HANDLE  javaProcess;            /* Handle of the Java process. */
+    int     monitorLaunched;        /* TRUE to monitor the launched process, FALSE to nonitor the Java process. */
 #else
     pid_t   wrapperPID;             /* PID of the Wrapper process. */
     pid_t   javaPID;                /* PID of the Java process. */
 #endif
+    int     monitorRedirectLogLevel;/* The log level to use when warning about the monitored process redirection. */
     int     wState;                 /* The current state of the wrapper */
     int     jState;                 /* The current state of the jvm */
     TICKS   jStateTimeoutTicks;     /* Tick count until which the current jState is valid */
@@ -390,8 +406,9 @@ struct WrapperConfig {
     int     restartRequested;       /* WRAPPER_RESTART_REQUESTED_NO, WRAPPER_RESTART_REQUESTED_AUTOMATIC, or WRAPPER_RESTART_REQUESTED_CONFIGURED if the another JVM should be launched after the current JVM is shutdown. Only set if exitRequested is set. */
     int     shutdownActionTriggered; /* TRUE if any action causing a JVM exit was triggered by the Wrapper (see the WRAPPER_ACTION_SOURCE_CODE_* definitions above).
                                      *  This flag indicates that the action specified by wrapper.on_exit should be ignored. It must be reset after the JVM is
-                                     *  restarted so that wrapper.on_exit can continue to catch JVM exit codes. */
+                                     *  restarted so that wrapper.on_exit can continue to catch JVM exit codes. This flag is only needed for actions being triggered after the JVM was started. */
     TCHAR*  shutdownActionPropertyName; /* Keep the name of the property that originated the shudown for logging. */
+    int     stopPacketReceived;     /* TRUE if a stop message was received, needed to be able to prevent sending other messages after that point. */ 
     int     stoppedPacketReceived;  /* TRUE if the STOPPED packet was received before a restart. */
     int     restartPacketReceived;  /* TRUE if the RESTART packet was received before a restart. */
     int     jvmRestarts;            /* Number of times that a JVM has been launched since the wrapper was started. */
@@ -446,6 +463,7 @@ struct WrapperConfig {
     gid_t   statusFileGroup;        /* Group to use when creating the status file. */
     gid_t   javaStatusFileGroup;    /* Group to use when creating the java status file. */
     gid_t   anchorFileGroup;        /* Group to use when creating the anchor file. */
+    int     javaNewProcessGroup;    /* Indicates whether the Java process should be created in a new process group or not (Unix only). */
 #endif
     int     ignoreSignals;          /* Mask that determines where the Wrapper should ignore any catchable system signals.  Can be ingored in the Wrapper and/or JVM. */
     TCHAR   *consoleTitle;          /* Text to set the console title to. */
@@ -467,6 +485,7 @@ struct WrapperConfig {
     DWORD   ntServicePriorityClass; /* Priority at which the Wrapper and its JVMS will run.
                                      * {HIGH_PRIORITY_CLASS | IDLE_PRIORITY_CLASS | NORMAL_PRIORITY_CLASS | REALTIME_PRIORITY_CLASS} */
     TCHAR   *ntServiceAccount;      /* Account name to use when running as a service.  NULL to use the LocalSystem account. */
+    int     ntServiceAddLogonAsService; /* Automatically add the 'Log on as a service' privilege to the service account. */
     TCHAR   *ntServicePassword;     /* Password to use when running as a service.  NULL means no password. */
     int     ntServicePrompt;        /* If true then the user will be prompted for a account name, domain,  password when installing as a service. */
     int     ntServicePasswordPrompt; /* If true then the user will be prompted for a password when installing as a service. */
@@ -497,7 +516,7 @@ struct WrapperConfig {
     int     signalHUPMode;          /* Controls what happens when the Wrapper receives a HUP signal. */
     int     signalUSR1Mode;         /* Controls what happens when the Wrapper receives a USR1 signal. */
     int     signalUSR2Mode;         /* Controls what happens when the Wrapper receives a USR2 signal. */
-    int     jvmStopped;             /* Flag which remembers the the stopped state of the JVM process. */
+    int     jvmStopped;             /* Flag which remembers the stopped state of the JVM process. */
 #endif
     
     int pendingPingQueueOverflow;   /* Flag which is set to true if the PendingPingQueue overflows the limit of WRAPPER_MAX_PENDING_PINGS. */
@@ -525,14 +544,15 @@ struct WrapperConfig {
  #endif
     int     ctrlCodeDumpTrapped;    /* The configured thread dump control code was trapped. */
 #else
-    int     signalInterruptTrapped; /* SIGINT was trapped. */
-    int     signalQuitTrapped;      /* SIGQUIT was trapped. */
-    int     signalQuitKernel;       /* SIGQUIT was originated by the kernel. */
-    int     signalChildTrapped;     /* SIGCHLD was trapped. */
-    int     signalTermTrapped;      /* SIGTERM was trapped. */
-    int     signalHUPTrapped;       /* SIGHUP was trapped. */
-    int     signalUSR1Trapped;      /* SIGUSR1 was trapped. */
-    int     signalUSR2Trapped;      /* SIGUSR2 was trapped. */
+    volatile sig_atomic_t signalInterruptTrapped; /* SIGINT was trapped. */
+    volatile sig_atomic_t signalQuitTrapped;      /* SIGQUIT was trapped. */
+    volatile sig_atomic_t signalQuitSkip;         /* SIGQUIT should be ignored. */
+    volatile sig_atomic_t signalChildTrapped;     /* SIGCHLD was trapped. */
+    volatile sig_atomic_t signalChildContinuedTrapped; /* SIGCHLD with CLD_CONTINUED was trapped. */
+    volatile sig_atomic_t signalTermTrapped;      /* SIGTERM was trapped. */
+    volatile sig_atomic_t signalHUPTrapped;       /* SIGHUP was trapped. */
+    volatile sig_atomic_t signalUSR1Trapped;      /* SIGUSR1 was trapped. */
+    volatile sig_atomic_t signalUSR2Trapped;      /* SIGUSR2 was trapped. */
 #endif
 };
 
@@ -553,6 +573,7 @@ struct WrapperConfig {
 #define WRAPPER_MSG_START_PENDING (char)105
 #define WRAPPER_MSG_STARTED       (char)106
 #define WRAPPER_MSG_STOPPED       (char)107
+#define WRAPPER_MSG_JAVA_PID      (char)108
 #define WRAPPER_MSG_KEY           (char)110
 #define WRAPPER_MSG_BADKEY        (char)111
 #define WRAPPER_MSG_LOW_LOG_LEVEL (char)112
@@ -578,7 +599,7 @@ struct WrapperConfig {
 
 extern WrapperConfig *wrapperData;
 extern Properties    *properties;
-
+extern int handleSignals;
 extern TCHAR wrapperClasspathSeparator;
 
 /**
@@ -727,12 +748,15 @@ extern struct tm wrapperGetBuildTime();
 #ifdef WIN32
 extern int initializeStartup();
 extern void disposeStartup();
+#else
+extern void closeStdinPipe();
+extern void disposeJavaIN();
 #endif
 extern void disposeJavaIO();
 extern int initializeTimer();
 extern void disposeTimer();
 
-extern int showHostIds(int logLevel);
+extern int showHostIds(int logLevel, int notUsed);
 #ifdef WIN32
 /**
  * Returns TRUE if the Wrapper process is associated with a console or if one
@@ -796,8 +820,8 @@ extern int wrapperBuildJavaClasspath(TCHAR **classpath);
 extern int wrapperBuildJavaCommandArray(TCHAR ***strings, int *length, int addQuotes, const TCHAR *classpath);
 extern void wrapperFreeStringArray(TCHAR **strings, int length);
 
-extern int wrapperInitialize(int silent);
-extern void wrapperDispose();
+extern int wrapperInitialize();
+extern void wrapperDispose(int exitCode);
 
 /**
  * Output the version.
@@ -898,6 +922,8 @@ extern void wrapperCheckConsoleWindows();
  */
 extern void verifyEmbeddedSignature();
 
+extern TCHAR* getExceptionName(DWORD exCode, int nullOnUnknown);
+
 extern int exceptionFilterFunction(PEXCEPTION_POINTERS exceptionPointers);
 BOOL isSecondary();
 BOOL extern elevateThis(int argc, TCHAR **argv);
@@ -913,10 +939,12 @@ int wrapperServiceStatus(const TCHAR* serviceName, const TCHAR* serviceDisplayNa
 extern int wrapperRemove();
 #define OSBUFSIZE 256
 BOOL GetOSDisplayString(TCHAR** pszOS);
+void wrapperCheckMonitoredProcess(DWORD javaPID);
 #else
+void wrapperCheckMonitoredProcess(pid_t javaPID);
 extern void changePidFileGroup(const TCHAR* filename, gid_t newGroup);
 extern void wrapperMaintainSignals();
-extern TCHAR* findPathOf(const TCHAR *exe, const TCHAR* name);
+extern TCHAR* findPathOf(const TCHAR *exe, const TCHAR* name, int compat);
 #endif
 
 /**
@@ -924,15 +952,7 @@ extern TCHAR* findPathOf(const TCHAR *exe, const TCHAR* name);
  */
 extern int wrapperInitializeRun();
 
-/**
- * Cause the current thread to sleep for the specified number of milliseconds.
- *  Sleeps over one second are not allowed.
- *
- * @param ms Number of milliseconds to wait for.
- *
- * @return TRUE if the was interrupted, FALSE otherwise.  Neither is an error.
- */
-extern int wrapperSleep(int ms);
+extern void wrapperSleep(int ms);
 
 /**
  * Reports the status of the wrapper to the service manager
@@ -960,7 +980,7 @@ extern int wrapperReadChildOutputBlock(char *blockBuffer, int blockSize, int *re
  * Checks on the status of the JVM Process.
  * Returns WRAPPER_PROCESS_UP or WRAPPER_PROCESS_DOWN
  */
-extern int wrapperGetProcessStatus(TICKS nowTicks, int sigChild);
+extern int wrapperGetProcessStatus(TICKS nowTicks, int childContinued);
 
 /**
  * Pauses before launching a new JVM if necessary.
@@ -1160,7 +1180,23 @@ extern int wrapperBuildJavaVersionCommand();
  *
  * @return The final index into the strings array, or -1 if there were any problems.
  */
-extern int wrapperBuildJavaCommandArrayJavaCommand(TCHAR **strings, int addQuotes, int detectDebugJVM, int index);
+extern int wrapperBuildJavaCommandArrayJavaCommand(TCHAR **strings, int addQuotes, int index, int shallow);
+
+/**
+ * The Java version will not parse correctly if the Java command is set to 'jdb' or 'javaw'.
+ *  - jdb has a different output when the '-version' parameters is specified.
+ *  - javaw has no std output.
+ *  This function tries to resolve the command used to request the Java version by searching
+ *  for the 'java' command located in the same directory, if it exists. If it doesn't, or if
+ *  the directory can't be resolved, a warning will be display and execution will continue to
+ *  let the Java version be resolved to its default value. 
+ *
+ * @param pCommand pointer to the Java command (the value may be modified).
+ * @param addQuotes TRUE if the resolved command should be surrounded by quotes, FALSE otherwise.
+ *
+ * @return TRUE if we ran out of memory, FALSE otherwise. 
+ */
+extern int wrapperResolveJavaVersionCommand(TCHAR **pCommand, int addQuotes);
 
 /**
  * Build the java command line.
@@ -1218,9 +1254,10 @@ extern TICKS wrapperAddToTicks(TICKS start, int seconds);
  * If there are any problems then a non-zero value will be returned.
  *
  * @param dir Directory to change to.
- * @param logErrors TRUE if errors should be logged.
+ *
+ * @return TRUE if the directory failed to be set, FALSE otherwise.
  */
-extern int wrapperSetWorkingDir(const TCHAR* dir, int logErrors);
+extern int wrapperSetWorkingDir(const TCHAR* dir);
 
 /******************************************************************************
  * Protocol callback functions
@@ -1258,7 +1295,7 @@ extern void wrapperStartPendingSignaled(int waitHint);
 extern void wrapperStartedSignaled();
 
 #ifdef WIN32
-int wrapperCheckPPid(DWORD pid, DWORD ppid);
+int wrapperCheckPPid(DWORD pid, DWORD ppid, int depth, int* pError);
 #endif
 
 int checkPidFile();
